#ifndef NETGEN_VERSION_HPP_INCLUDED
#define NETGEN_VERSION_HPP_INCLUDED
#define NETGEN_VERSION "6.2.2505"
#define NETGEN_VERSION_MAJOR 6
#define NETGEN_VERSION_MINOR 2
#define NETGEN_VERSION_PATCH 2505
#define NETGEN_VERSION_TWEAK 0
#define NETGEN_VERSION_HASH "g41893720"
#endif // NETGEN_VERSION_HPP_INCLUDED
