#ifndef NETGEN_CONFIG_HPP_INCLUDED___
#define NETGEN_CONFIG_HPP_INCLUDED___

#define NETGEN_USE_NATIVE_ARCH          0
#define NETGEN_USE_GUI                  1
#define NETGEN_USE_PYTHON               1
#define NETGEN_USE_MPI                  1
#define NETGEN_USE_MPI4PY               0
#define NETGEN_USE_OCC                  1
#define NETGEN_USE_JPEG                 1
#define NETGEN_USE_MPEG                 1
#define NETGEN_USE_CGNS                 1
#define NETGEN_USE_NUMA                 1
#define NETGEN_INTEL_MIC                1
#define NETGEN_INSTALL_PROFILES         0
#define NETGEN_USE_CCACHE               1
#define NETGEN_USE_INTERNAL_TCL         1
#define NETGEN_ENABLE_UNIT_TESTS        1
#define NETGEN_ENABLE_CPP_CORE_GUIDELINES_CHECK 0
#define NETGEN_USE_SPDLOG               0
#define NETGEN_DEBUG_LOG                0
#define NETGEN_USE_CHECK_RANGE          0
#define NETGEN_BUILD_STUB_FILES         0
#define NETGEN_BUILD_FOR_CONDA          0
#define NETGEN_SHARED_LIBRARY_SUFFIX    ".so"

#endif // NETGEN_CONFIG_HPP_INCLUDED___
