#ifndef FILE_GPRIM
#define FILE_GPRIM

/* *************************************************************************/
/* File:   gprim.hpp                                                        */
/* Author: Joachim Schoeberl                                               */
/* Date:   14. Aug. 97                                                     */
/* *************************************************************************/


#include <myadt.hpp>
#include <linalg.hpp>




#include "geomobjects.hpp"
#include "geomops.hpp"
#include "geomfuncs.hpp"

#include "geom2d.hpp"
#include "geom3d.hpp"

#include "geomtest3d.hpp"
#include "transform3d.hpp"

#include "adtree.hpp"

#include "spline.hpp"
#include "splinegeometry.hpp"


#endif
