#ifndef VISUAL_API_HPP_INCLUDED
#define VISUAL_API_HPP_INCLUDED

#ifdef nggui_EXPORTS
#define NGGUI_API NGCORE_API_EXPORT
#else
#define NGGUI_API NGCORE_API_IMPORT
#endif

#endif // VISUAL_API_HPP_INCLUDED
