#ifndef FILE_LINALG
#define FILE_LINALG

/* *************************************************************************/
/* File:   linalg.hpp                                                      */
/* Author: Joachim Schoeberl                                              */
/* Date:   01. Oct. 94                                                    */
/* *************************************************************************/

/* 

   Data types for basic linear algebra
   
   The basic concepts include the data types 
   
    Vector
    SparseMatrix
    DenseMatrix

*/


#include "../include/myadt.hpp"
#include "vector.hpp"
#include "densemat.hpp"
#include "polynomial.hpp"
#endif


